{- | This module provides commonly used computations on elliptic curves.

   In the following, assume the elliptic curve of interest is

   y^2 = x^3 + ax + b (mod n)
-}
module EllipticCurve(valid, ecm, op, power, power_proj)  where

import Data.Bits

import ExtendedEuclid

{- | @valid n a b@ checks if y^2 = x^3 + ax + b is a valid
   elliptic curve mod n, i.e., the discrimant 4a^3+27b relatively prime to n.
   If valid, returns Nothing; if invalid, returns the gcd.
-}
valid :: Integer -> Integer -> Integer -> Maybe Integer
valid n a b = if g==1 then Nothing else Just g
    where g = (4*a^3 + 27*b*b) `gcd` n

{- | @op n a b (x1,y1) (x2,y2)@ computes the elliptic curve operation
   on (x1,y1), (x2,y2). (Need to say more.)
-}
op :: Integer -> Integer -> Integer -> (Integer,Integer) -> (Integer,Integer)
   -> Either Integer (Integer,Integer)
op n a b s@(x,y) t@(x',y') = if g == 1 then Right (xo, (m*(x-xo) - y) `rem` n) else Left g
    where xo = (m*m - x - x') `rem` n
          (m,zo) | s/=t = ((y-y') * rzo, x-x')
                 | otherwise = ((3*x*x + a) * rzo, 2*y)
          (rzo, g, _, _) = pinverse n zo

{- | @power n a b (x,y) k@ raises (x,y) to the power k: repeats the elliptic curve
   operation k times. (Need to say more.)
-}
power :: (Bits k) => Integer -> Integer -> Integer -> (Integer,Integer) -> k
      -> Either Integer (Integer,Integer)
power n _ _ _ 0 = Left n
power n a b s e = go s s (e-1)
    where
      -- go s t e = s*t^e
      go s _ 0 = Right s
      go s t 1 = s**t
      go s t e | e `testBit` 0 = s**t >>= \s' -> t**t >>= \t' -> go s' t' (e `shiftR` 1)
               | otherwise = t**t >>= \t' -> go s t' (e `shiftR` 1)
      (**) = op n a b

ecm0 (a,b,s) max n = case valid n a b of
  Just g -> (g, n `quot` g)
  Nothing -> (g, n `quot` g)
    where g = gcd n z
          Left z = search s 1
          search t i | i > max = Left n
                     | otherwise = t ^^ i >>= \t' -> search t' (i+1)
          (^^) = power n a b

ecmsearch (a,b,s) max n = search s 1
    where
          search t i | i > max = []
                     | otherwise = case t ^^ i of Left a -> [Left a]
                                                  Right b -> Right b : search b (i+1)
          (^^) = power n a b

{- | @power_proj n a b (x,z) k@ raises (x,z) to the power k: repeats the elliptic curve
   operation k times.

   This function is in terms of projective points [x,y,z] representing (x\/z,y\/z) or
   the identity point or degenerate points (z not relatively prime to n).
   Furthermore the y coordinate is omitted, leading to just (x,z).

   Furthermore x is assumed non-zero mod n.
-}
power_proj :: (Bits k) => Integer -> Integer -> Integer -> (Integer,Integer) -> k
           -> (Integer,Integer)
power_proj _ _ _ _ 0 = (0,0)
power_proj n a b s@(x0,z0) i = fst (go i)
    where
      -- "squares" a point, i.e., (x,z)*(x,z). So if (x,z)=s^j, then get s^(2j).
      sq (x,z) = let xsq = x*x `rem` n
                     xcb = xsq*x `rem` n
                     zsq = z*z `rem` n
                     zcb = zsq*z `rem` n
                     azsq = a*zsq `rem` n
                     bzcb = b*zcb `rem` n
                     u = xsq - azsq
                 in ( (u*u - 8*x*bzcb) `rem` n,
                      4*z*(xcb + x*azsq + bzcb) `rem` n )
      -- assuming (x,z)=s^j and (x',z')=s^(j+1), compute s^(2j+1)
      -- note: depends also on original s=(x0,z0)
      op (x,z) (x',z') = let xx' = x*x' `rem` n
                             zz' = z*z' `rem` n
                             xz' = x*z' `rem` n
                             x'z = x'*z `rem` n
                             u = xx' - a*zz'
                             v = x'z - xz'
                         in ( z0*(u*u - 4*b*zz'*(xz' + x'z)) `rem` n,
                              x0*v*v `rem` n )
      -- go i computes (s^i, s^(i+1))
      go 1 = (s, sq s)
      go i | i `testBit` 0 = (op t t', sq t')
           | otherwise = (sq t, op t t')
           where (t, t') = go (i `shiftR` 1)

{- | Factoring method based on elliptic curves.

   @ecm (a,b,(x,y)) max n@ tries to factor n by looking for degeneracy in

   (...(((x,y)^1)^2)^...)^k

   for k up to max.

   Assuming x is non-zero mod n.

   Validity test of the given elliptic curve is performed too. If invalid,
   returns the resulting factoring.
-}
ecm :: [Int] -> (Integer,Integer,(Integer,Integer)) -> Int -> Integer -> (Integer,Integer)
ecm primes (a,b,(x,_)) max n = case valid n a b of
  Just g -> (g, n `quot` g)
  Nothing -> search (powers (x,1) lowprimes) -- use projective rep omitting y
    where
      lowprimes = takeWhile (max >=) primes
      powers _ [] = []
      powers s (p:ps) = case s^^(p ^ floor(logi max / logi p)) of si@(_,z) -> z : powers si ps
      (^^) = power_proj n a b
      logi n = log (fromIntegral n)
      search [] = (n,1)
      search s = case mysplitAt 16 s of
          (ps,x,ss)
              | g==1 -> search ss
              | g/=n -> (g, n `quot` g)
              | otherwise -> lookcloser ps
              where g = n `gcd` x
      lookcloser ps = let g = head (dropWhile (1 ==) (map (gcd n) ps)) in
                      seq g (g, n `quot` g)
{-
ecm :: (Integer,Integer,(Integer,Integer)) -> Int -> Integer -> (Integer,Integer)
ecm (a,b,(x,_)) max n = case valid n a b of
  Just g -> (g, n `quot` g)
  Nothing -> search (powers (x,1) 1) -- use projective rep omitting y
    where
      powers s i
          | i > max = []
          | otherwise = case s ^^ i of si@(_,z) -> z : powers si (i+1)
      (^^) = power_proj n a b
      search [] = (n,1)
      search s = case mysplitAt 16 s of
          (ps,x,ss)
              | g==1 -> search ss
              | g/=n -> (g, n `quot` g)
              | otherwise -> lookcloser ps
              where g = n `gcd` x
      lookcloser ps = let g = head (dropWhile (1 ==) (map (gcd n) ps)) in
                      seq g (g, n `quot` g)
-}

{-  mysplitAt n xs = (a, last a, b)
  = splitAt n xs = (a, b)
  assuming xs non-empty and n>=1.
-}
mysplitAt 1 (x:xs) = ([x], x, xs)
mysplitAt _ xs@(x:ns@[]) = (xs, x, ns)
mysplitAt n (x:xs) = (x:xs', x', xs'')
    where (xs', x', xs'') = mysplitAt (n-1) xs

-- ecmtest a max n = ecm (a,1,(0,1)) max n

instance Monad (Either a) where
    return b = Right b
    (Left a) >>= _ = Left a
    (Right b) >>= f = f b
    fail _ = Left undefined

