{- | This module provides the extended Euclidean algorithm
   and derived functions.
-}

module ExtendedEuclid(euclid, inverse, pinverse) where

{- | @euclid m n@ = (g, u, v) with

   u&#215;m + v&#215;n = g = gcd m n &#8805; 0
-}
euclid :: Integer -> Integer -> (Integer, Integer, Integer)
euclid a b = go 1 0 a 0 1 b
  where
  go u v w x y z | z==0 = if w>=0 then (w, u, v) else (-w, -u, -v)
                 | otherwise = seq u' $ seq v' $ go x y z u' v' w'
    where (q,w') = quotRem w z
          u' = u-q*x
          v' = v-q*y

test a b | a*x + b*y == g = e
         | otherwise = error "wrong"
    where e@(g, x, y) = euclid a b

{- | @inverse m b@ is the inverse of b (mod m) in 1..m-1
   if it exists (if gcd b m = 1).

   It throws @error \"inverse\"@ if not.

   Assumption: m>0, b unrestricted.
-}
inverse :: Integer -> Integer -> Integer
inverse m b | g == 1 = v `mod` m
            | otherwise = error "inverse"
    where (g, _, v) = euclid m b

{- | @pinverse@ tries to compute something like inverse even when
   it does not exist per se. (p for pseudo)

   @pinverse m b@ = (r, g, m\/g, b\/g) where

   * g = gcd m b

   * r is the inverse of b\/g (mod m\/g) in 1..m\/g-1

   Assumption: m>0, b unrestricted.
-}
pinverse :: Integer -> Integer -> (Integer,Integer,Integer,Integer)
pinverse m b = (v `mod` m', g, m', b')
    where (g, _, v) = euclid m b
          m' = m `quot` g
          b' = b `quot` g
