{- | Pollard's p-1 factoring method.

   Following Fermat's little theorem b^(p-1) - 1 = 0 (mod p)
   for b relatively prime to p, if p is a prime factor of n
   and furthermore p-1 divides k! for small k
   (implying prime factors of p-1 are small), then
   b^(k!) - 1 = 0 (mod p) and hopefully factors n.
   So we try k from 1 to a bound.
   Number of iterations is likely about the largest prime factor of p-1.
-}
module PollardP1(p_1) where

import PowerMod

{- | Pollard's p-1 factoring method.
-}
p_1 :: [Int]                    -- ^ list of small primes
    -> Integer                  -- ^ base to raise power from
    -> Int                      -- ^ maximum power tried
    -> Integer                  -- ^ n to be factored
    -> (Integer,Integer)        -- ^ good factoring (d,n\/d) or (n,1)
p_1 primes b max n = if gcd b n /= 1 then (gcd b n, n `quot` gcd b n)
                     else search (powers b lowprimes)
    where
      lowprimes = takeWhile (max >=) primes
      powers _ [] = []
      powers x (p:ps) = let xi = x ^^ (p ^ floor(logi max / logi p))
                        in (xi - 1) : powers xi ps
      (^^) = powermod n
      search [] = (n,1)
      search s = case mysplitAt 16 s of
          (ps,x,ss)
              | g==1 -> search ss
              | g/=n -> (g, n `quot` g)
              | otherwise -> lookcloser ps
              where g = n `gcd` x
      lookcloser ps = let g = head (dropWhile (1 ==) (map (gcd n) ps)) in
                      seq g (g, n `quot` g)

logi n = log (fromIntegral n)

{-  mysplitAt n xs = (a, last a, b)
  = splitAt n xs = (a, b)
  assuming xs non-empty and n>=1.
-}
mysplitAt 1 (x:xs) = ([x], x, xs)
mysplitAt _ xs@(x:ns@[]) = (xs, x, ns)
mysplitAt n (x:xs) = (x:xs', x', xs'')
    where (xs', x', xs'') = mysplitAt (n-1) xs

{- This nice combinator may be useful one day.

myfix f t x i = go x i
    where
      granularity = 16
      go x0 i0 = fast granularity x0 i0
          where
            fast 0 x i = case t x i of
                            Nothing -> go x i
                            Just (g,d) | d /= 1 -> (g,d) 
                                       | otherwise -> f slow x0 i0
            fast j x i = f (fast (j-1)) x i
      slow x i =
          case t x i of
            Just gd -> gd
            _ -> f slow x i

-}
