{- | Pollard's rho factoring method.

   Pick a quadratic f(x) = x^2+c mod n. Define x(i) = f^i(2).
   (The 2 is arbitrary and unimportant.)
   Hopefully this is a fairly random sequence eventually periodic
   modulo the smallest prime factor p of n, so that
   for some i,j, gcd(x(i) - x(j), n) is a good factor of n.
   Average cost is about 3 sqrt p iterations.
-}
module PollardRho(rho) where

import Data.List(foldl1')

import PowerMod

{- | Pollard's rho factoring method.
   Returns a factoring (d, n\/d) or (n,1).
-}
rho :: Integer              -- ^ determines the quadratic used: x^2 + this
    -> Int                  -- ^ try this many iterations (more or less)
    -> Integer              -- ^ the number n to be factored
    -> (Integer,Integer)    -- ^ hopefully a proper factoring, or else (n,1)
rho c max n = search (brentlist 2 1 (f c))
    where
      f x = (x*x + c) `rem` n
      max2 = max `quot` 2
      -- Sequence of x(i)-x(j) under Brent's accelerated cycle detection.
      brentlist x r y
          | r > max2 = []
          | otherwise = inner y 0
          where 
            inner y i
                | i/=r = let y' = f y in x-y' : inner y' (i+1)
                | otherwise = brentlist y (2*r) (app r f y)
      search [] = (n,1)
      search s = case splitAt 16 s of
          (ps,ss)
              | g==1 -> search ss
              | g/=n -> (g, n `quot` g)
              | otherwise -> lookcloser ps
              where g = n `gcd` foldl1' (\x y -> x*y `rem` n) ps
      lookcloser ps = let g = head (dropWhile (1 ==) (map (gcd n) ps)) in
                      seq g (g, n `quot` g)

{- note strictness -}
app n f x = go n x
    where go 0 x = x
          go n x = go (n-1) $! (f x)

