{- | This module uses trial division to factor a number
   completely with respect to a given list of primes.

   Module "SmallPrimes" is a great source of lists of primes.
-}
module TrialDivision(factor, efactor, factor') where

{- | Factors a positive number into non-decreasing factors.
   Multiplicity is reflected as multiple occurrences.
   The factoring is as complete as the given primes avail.
   The input prime list is assumed to be primes and increasing.
   Example:

   > factor [2,3,5] (2*3*3*49) = [2, 3, 3, 49]
-}
factor :: [Integer] -> Integer -> [Integer]
factor _ 1 = []
factor [] n = [n]
factor s@(p:ps) n | p*p > n   = [n]
                  | otherwise = reduce s n

reduce s@(p:ps) n | r == 0    = p : reduce s q
                  | otherwise = factor ps n
  where (q,r) = quotRem n p

{- | Factors a positive number into non-decreasing factors.
   Multiplicity is reflected as multiple occurrences.
   The factoring is as complete as the given primes avail;
   however, unlike 'factor' above, the part that cannot be
   completely factored is returned separately.
   The input prime list is assumed to be primes and increasing.
   Example:

   > factor' [2,3,5] (2*3*3*49) = ([2, 3, 3], Just 49)
   > factor' [2,3,5] (2*3*3*7) = ([2, 3, 3, 7], Nothing)
-}
factor' :: [Integer] -> Integer -> ([Integer], Maybe Integer)
factor' _ 1 = ([], Nothing)
factor' [] n = ([], Just n)
factor' s@(p:ps) n | p*p > n = ([n], Nothing)
                   | r /= 0 = factor' ps n
                   | otherwise = (p : more, res)
    where (q,r) = quotRem n p
          (more, res) = factor' s q

{- | Factors a positive number into non-decreasing factors.
   Multiplicity is reflected as pairs of (prime, multiplicity).
   The factoring is as complete as the given primes avail.
   The input prime list is assumed to be primes and increasing.
   Example:

   > efactor [2,3,5] (2*3*3*49) = [(2,1), (3,2), (49,1)]
-}
efactor :: [Integer] -> Integer -> [(Integer,Integer)]
efactor _ 1 = []
efactor [] n = [(n,1)]
efactor s@(p:ps) n | p*p > n   = [(n,1)]
                   | r == 0    = ereduce s q 1
                   | otherwise = efactor ps n
  where (q,r) = quotRem n p

ereduce s@(p:ps) n i | r == 0    = ereduce s q $! (i+1)
                     | otherwise = (p,i) : efactor ps n
  where (q,r) = quotRem n (p)
