{- Factorer with presets.
-}
module Main(main) where

import Debug.Trace

import Control.Exception(evaluate)
import Control.Monad(msum, replicateM, when)
import Data.List(sort)
import System.Environment(getArgs, getProgName)
import System.IO(hFlush, hPutStr, hPutStrLn, stdout, stderr)
import System.Random(randomRIO)

import EllipticCurve
import PollardP1
import PollardRho
import SmallPrimes
import StrongProbablePrime
import TrialDivision

{- Preset parameters -}
-- we have all small primes below this
primeBaseBound = 500000
-- strong probable prime test tries this many small primes
probPrimeStrength = 1000
-- Pollard's rho runs this many iterations for each x^2+c
rhoMax = 500000
-- the c's in x^2+c in Pollard's rho; a few random small primes also used
rhoBase = [1,2,3,5,7]
-- how many is "a few"
rhoRandom = 5
-- Pollard's p-1 runs this many iterations for each base
p1Max = 100000
-- the bases in Pollard's p-1; a few random numbers also used
p1Base = [2,3,5,7,11]
-- how many is "a few"
p1Random = 5

primeBase :: [Integer]
primebase :: [Int]
(primebase,primeBase,_) = smallprimes primeBaseBound

tryPrimeBase :: Integer -> ([Integer], Maybe Integer)
tryPrimeBase n = factor' primeBase n

probPrime :: Integer -> Bool
probPrime n = all (spp n) (take probPrimeStrength primeBase)

apply :: [Int -> Integer -> (Integer,Integer)] -> Int -> Integer -> Maybe (Integer,Integer)
apply fs m n = msum (map run fs)
    where run f = case f m n of (_,1) -> Nothing
                                a -> Just a

apply2 :: [IO a] -> (a -> Int -> Integer -> (Integer,Integer)) -> Int -> Integer
       -> IO (Maybe (Integer,Integer))
apply2 [] _ _ _ = return Nothing
apply2 (x:xs) f m n = do
  v <- x
  case f v m n of
    (_,1) -> apply2 xs f m n
    a -> return (Just a)

tryRho :: Integer -> IO (Maybe (Integer,Integer))
tryRho n = do bs <- replicateM rhoRandom
                    ((primeBase !!) `fmap` randomRIO (0, length primeBase - 1))
              evaluate (apply (map rho (rhoBase ++ bs)) rhoMax n)

tryP1 :: Integer -> IO (Maybe (Integer,Integer))
tryP1 n = do bs <- replicateM p1Random (randomRIO (2, 2^30))
             evaluate (apply (map (p_1 primebase) (p1Base ++ bs)) p1Max n)

tryEC :: Integer -> IO (Maybe (Integer,Integer))
tryEC n = apply2 (repeat randomParamIO) (ecm primebase) ecmax n
    where r = randomRIO (1, 2^30)
          randomParamIO = do
            a <- r
            x <- r
            y <- r
            let b = (y^2 - x^3 - a*x) `rem` n
            return (a,b,(x,y))
          ecmax = ceiling (1.1 * exp (sqrt (0.5 * log x * log (log x))))
          x = sqrt (fromIntegral n)

tryEverything :: Integer -> IO ([Integer],[Integer])
tryEverything n = do
  ePutStrLn ("Trying to factor " ++ show n)
  let (ps, res) = tryPrimeBase n
  ePutStrLn ("Small prime factors: " ++ unwords (map show ps))
  case res of
    Nothing -> return (ps, [])
    Just n' -> do
      (bps, cs) <- bigGuns n' [("rho", tryRho), ("p-1", tryP1), ("ec", tryEC)]
      return (ps ++ sort bps, sort cs)

bigGuns n guns = do
  ePutStr ("Bignum: " ++ show n ++ " ")
  if probPrime n
    then ePutStrLn "is prime" >> return ([n], [])
    else thru n guns

thru n [] = ePutStrLn "fail" >> return ([], [n])
thru n gs@((name,method):more) = do
  ePutStr (name ++ " ")
  a <- method n
  case a of
    Nothing -> thru n more
    Just (x,y) -> do
      ePutStrLn ("= " ++ show x ++ " * " ++ show y)
      (xps, xcs) <- bigGuns x gs
      (yps, ycs) <- bigGuns y gs
      return (xps++yps, xcs++ycs)

main = do
  as <- getArgs
  case as of
    [a0] -> do
      let n = read a0
      (ps, cs) <- tryEverything n
      putStrLn (show n ++ " factors:")
      when (not (null ps))
           (putStrLn (unwords (map show ps) ++ " primes"))
      when (not (null cs))
           (putStrLn (unwords (map show cs) ++ " composites"))
    _ -> do p <- getProgName
            putStrLn (p ++ " number")

ePutStr s = hPutStr stderr s
ePutStrLn s = hPutStrLn stderr s

